/***********************************************************************************
* Program Name : buildvol
* Description  : <description>
* Author       : Guillaume Major
* Date         : August 8, 2006
* License      : Freeware. Do whatever you want with it!
***********************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <direct.h>
#include <windows.h>
#include "lzw.h"

/* Resource types */
#define RES_LOGIC	0
#define RES_PICTURE	1
#define RES_VIEW	2
#define RES_SOUND	3

char *ResTypes[] = { "logic", "picture", "view", "sound" };
char *ResFiles[] = { "logDir", "picDir", "viewDir", "sndDir" };

//#pragma pack(1)

/* VOL directory entry structure */
struct DIRECTORY_ENTRY {
	unsigned char	byte1;
	unsigned char	byte2;
	unsigned char	byte3;
};

/* Resource header */
struct RESOURCE_HEADER {
	unsigned short	resId;
	unsigned char	volFile;
	unsigned short	size;
};

/* Resource header  v3 */
struct RESOURCE_HEADER_V3 {
	unsigned short	resId;
	unsigned char	volFile;
	unsigned short	originalSize;
	unsigned short	compressedSize;
};

/* V3 Directory File Header */
struct DIRECTORY_FILE_HEADER {
	unsigned short	logOffset;
	unsigned short	picOffset;
	unsigned short	viewOffset;
	unsigned short	sndOffset;
};

/* resource options */
struct RESOURCE_OPTIONS {
	int		compression;
};

/* Global variables */
char sResPath[MAX_PATH] = "";
char sDirPath[MAX_PATH] = "";
char sOutputPath[MAX_PATH] = "";
int  iAGIVer = 2;
int  compression = 0;
int  bigfile = 0;
int  volno = 0;

/* Function prototypes */
void ProcessDirFile(int resType);
unsigned int WriteResource(unsigned char *sResFile, unsigned int iVolFile, 
						   int iResType, struct RESOURCE_OPTIONS *resOptions);
void GetResourceOptions(unsigned char *sResFile, struct RESOURCE_OPTIONS *resOptions);
void DecryptMessageSection(char *buffer, int len);
unsigned int CompressPicture(char *buffer, char *compressed, unsigned int iResSize);
unsigned int GetVolSize(unsigned int iVolFile);
void CleanUp();
void ShowUsage();

/* Main */
int main(int argc, char **argv) {

	int i = 0;

	printf("\n");
	printf("AGI VOL file builder\n");
	printf("Version 1.2\n\n");

	/* show usage if no argument */
	if (argc == 1)
	{
		ShowUsage();
	}

	/* Get command line arguments */
	while (++i < argc && argv[i][0] == '-')
	{
		if (strcmp(argv[i], "-v3") == 0 ||
			strcmp(argv[i], "-V3") == 0)
		{
			iAGIVer = 3;
		}
		else if (strcmp(argv[i], "-c") == 0 ||
			strcmp(argv[i], "-C") == 0)
		{
			compression = 1;
		}
		else if (strcmp(argv[i], "-b") == 0 ||
			strcmp(argv[i], "-B") == 0)
		{
			bigfile = 1;
		}
		else
		{
			printf("Invalid option: %s\n\n", argv[i]);
			ShowUsage();
		}
	}

	if (iAGIVer == 2 && compression == 1)
	{
		printf("Can't use compression with version 2! Use with -v3 option.\n\n");
		ShowUsage();
	}
	
	if (argc-i < 3) {
		printf("No enough parameters\n\n");
		ShowUsage();
	}
	
	strncpy(sResPath, argv[i], MAX_PATH); 
	strncpy(sDirPath, argv[i+1], MAX_PATH); 
	strncpy(sOutputPath, argv[i+2], MAX_PATH); 

	/* Clean up the output directory */
	CleanUp();

	/* Process directory files */
	ProcessDirFile(RES_LOGIC);
	ProcessDirFile(RES_PICTURE);
	ProcessDirFile(RES_VIEW);
	ProcessDirFile(RES_SOUND);
						   								
	printf("Done!\n\n");

	/* Exit normally */	
	exit(0);
}

/***************************************************************************
** ProcessDirFile
**
** Purpose: Loops through the specified directory file to write resources in
** VOL files.
**
** Parameters:
**   iResType: The type of resource to process.
***************************************************************************/
void ProcessDirFile(int iResType) {

	FILE *dirFile, *newDirFile;
	struct DIRECTORY_ENTRY dir, newDir;
	unsigned int iResNum, iVolFile, iPos;
	unsigned char sDirFile[MAX_PATH];
	unsigned char sNewDirFile[MAX_PATH];
	unsigned char sResFile[MAX_PATH];
	struct RESOURCE_OPTIONS resOptions;

	printf("\nProcessing %ss\n\n", ResTypes[iResType]);

	sprintf(sDirFile, "%s\\%s", sDirPath, ResFiles[iResType]);
	sprintf(sNewDirFile, "%s\\%s", sOutputPath, ResFiles[iResType]);

	/* Open the original dir file for reading */
	if ((dirFile = fopen(sDirFile, "rb")) == NULL) {
		printf("Error opening directory file %s", sDirFile);
		exit(1);
	}

	/* Open the new dir file for writing */
	if ((newDirFile = fopen(sNewDirFile, "wb")) == NULL) {
		printf("Error opening new directory file %s", sNewDirFile);
		exit(1);
	}

	iResNum = 0;

	/* Loop through the directory file */
	while (!ferror(dirFile) && !feof(dirFile) && fread(&dir, sizeof(dir), 1, dirFile)) {
		memset(&newDir, 0xFF, sizeof(newDir));
		/* Verify if the resource exists */
		printf("%s.%i\n", ResTypes[iResType], iResNum);
		if (dir.byte1 != 0xFF) {
			/* get the vol file number */
			iVolFile = (dir.byte1 >> 4);
			sprintf(sResFile, "%s\\%s.%i", sResPath, ResTypes[iResType], iResNum);
			/* Get resource options */
			GetResourceOptions(sResFile, &resOptions);
			/* Write the resource in the VOL file */
			iPos = WriteResource(sResFile, iVolFile, iResType, &resOptions);
			if (iPos != 0xFFFFFFFF) {
				newDir.byte1 = ((iVolFile << 4) | (iPos & 0x0F0000) >> 16);
				newDir.byte2 = (iPos & 0x00FF00) >> 8;
				newDir.byte3 = (iPos & 0x0000FF);
			}
		}
		fwrite(&newDir, sizeof(newDir), 1, newDirFile);
		iResNum++;
	}

	/* Close files */
	fclose(dirFile);
	fclose(newDirFile);

}

/***************************************************************************
** WriteResource
**
** Purpose: Adds a resource to a VOL files.
**
***************************************************************************/
unsigned int WriteResource(unsigned char *sResFile, unsigned int iVolFile, 
						   int iResType, struct RESOURCE_OPTIONS *resOptions) {

	FILE *resFile, *volFile;
	unsigned char sVolFile[MAX_PATH];
	unsigned short iResSize;
	char acBuffer[65536];
	struct RESOURCE_HEADER header;
	struct RESOURCE_HEADER_V3 header_v3;
	fpos_t pos;

	unsigned short iCompSize;
	char acCompressed[65536];

	/* Open the resource file for reading */
	if ((resFile = fopen(sResFile, "rb")) == NULL) {
		printf("Error opening resource file %s\n", sResFile);
		getchar();
		return 0xFFFFFFFF;
	}	

	/* Read the entire resource file into the buffer */
	fseek(resFile, 0, SEEK_SET);
	iResSize = fread(acBuffer, sizeof(unsigned char), sizeof(acBuffer), resFile);

	/* verify if all resources must be written in one big file */
	if (bigfile) {
		/* increment vol file number if the current vol file is too big */
		if (GetVolSize(volno)+iResSize >= 1048576) {
			volno++;
		}
		iVolFile = volno;
	}

	/* Get the VOL file name */
    sprintf(sVolFile, "%s\\vol.%i", sOutputPath, iVolFile);

	/* Open the VOL file for appending */
	if ((volFile = fopen(sVolFile, "a+b")) == NULL) {
		printf("Error opening VOL file %s\n", sVolFile);
		getchar();
		return 0xFFFFFFFF;
	}

	/* Get the position of the cursor */
	fseek(volFile, 0, SEEK_END);
	fgetpos(volFile, &pos);
	
	if (iAGIVer == 2) {
		/* Build a v2 header */
		header.resId = 0x3412;
		header.volFile = iVolFile;
		header.size = iResSize;
			
		/* Write the header to the VOL file */
		fwrite(&header, sizeof(header), 1, volFile);

		/* Write the contents of the buffer to the VOL file */
		fwrite(acBuffer, sizeof(char), iResSize, volFile);
	}
	else {
		/* Check if the resource must be compressed */
		if (compression == 1 && resOptions->compression) {
			/* Compress the resource */
			memset(acCompressed, 0, sizeof(acCompressed));
			/* Is it a picture? */
			if (iResType == RES_PICTURE) {
				iCompSize = CompressPicture(acBuffer, acCompressed, iResSize);
			}
			else 
			{
				if (iResType == RES_LOGIC)
				{
					// Decrypt the message section before compression
					DecryptMessageSection(acBuffer, iResSize);
				}
				iCompSize = LZW_compress(acBuffer, acCompressed, iResSize);
			}
		}
		else
		{
			iCompSize = iResSize;
		}

		/* Build a v3 header */
		header_v3.resId = 0x3412;
		header_v3.volFile = iVolFile;
		header_v3.originalSize = iResSize;
		header_v3.compressedSize = iCompSize;

		/* if the resource is compressed and is a picture set the picture bit */
		if (iResSize != iCompSize && iResType == RES_PICTURE) {
			header_v3.volFile |= 0x80;
		}

		/* Write the header to the VOL file */
		fwrite(&header_v3, sizeof(header_v3), 1, volFile);

		if (iResSize != iCompSize) {
			/* Write the compressed resource to the VOL file */
			fwrite(acCompressed, sizeof(char), iCompSize, volFile);
		}
		else {
			/* Write the uncompressed resource to the VOL file */
			fwrite(acBuffer, sizeof(char), iResSize, volFile);
		}
	}

	/* Close files */
	fclose(volFile);
	fclose(resFile);

	return (unsigned int)pos;
}

/***************************************************************************
** GetResourceOptions
**
** Purpose: Gets the optional parameters for the specified resource.
**
***************************************************************************/
void GetResourceOptions(unsigned char *sResFile, struct RESOURCE_OPTIONS *resOptions)
{
	FILE *resOptFile;
	char sResOptFile[MAX_PATH];
	char line[256];
	char option[256] = "";
	char value[256] = "";
	char *pos;
	unsigned int i;

	/* Set default resource options */
	resOptions->compression = 1;

	sprintf(sResOptFile, "%s.options", sResFile);
	
	/* Check if the resource has an option file */
	if ((resOptFile = fopen(sResOptFile, "r")) != NULL) {
		/* Parse the option file */
		while (fgets(line, 255, resOptFile) != 0)
		{
			if ((pos = strstr(line, "=")) != NULL)
			{
				/* get the option and value */
				memcpy(option, line, pos-line);
				strcpy(value, pos+1);
				value[strlen(value)-1] = 0;
				/* lowercase the option */
				for (i=0; i<strlen(option); i++)
				{
					option[i] = tolower(option[i]);
				}
				/* compression option */
				if (strcmp(option, "compression") == 0)
				{
					resOptions->compression = atoi(value);
					if (resOptions->compression == 0)
					{
						printf("Compression disabled\n");
					}
				}
			}
		}
		fclose(resOptFile);
	}
}

/***************************************************************************
** CompressPicture
**
** Purpose: Compresses a picture resource.
**
***************************************************************************/
unsigned int CompressPicture(char *buffer, char *compressed, unsigned int iResSize)
{
	unsigned char *code = buffer;
	unsigned char *end = buffer+iResSize;
	unsigned int iCompSize = 0;
	unsigned char output_bits = 0;
	unsigned char input_bits = 0;
	unsigned char trim = 0;

	while (code < end)
	{
		if (trim == 1) {
			input_bits = 4;
			trim = 0;
		}
		else
		{
			input_bits = 8;
		}
		if (*code == 0xF0 || *code == 0xF2) {
			trim = 1;						
		}
		if (output_bits == 0 && input_bits == 8) {
			compressed[iCompSize++] = *code;
		}
		else if (output_bits == 0 && input_bits == 4)
		{
			compressed[iCompSize] = *code << 4;
			output_bits = 4;
		}
		else if (output_bits == 4 && input_bits == 4)
		{
			compressed[iCompSize++] |= *code & 0x0F;
			output_bits = 0;
		}
		else if (output_bits == 4 && input_bits == 8)
		{
			compressed[iCompSize++] |= *code >> 4;
			compressed[iCompSize] = (*code & 0x0F) << 4;
		}
		code++;		
	}
	
	if (output_bits == 4) {
		iCompSize++;
	}

	return iCompSize;
}

/***************************************************************************
** GetVolSize
**
** Purpose: Gets the size of a volume file.
**
***************************************************************************/
unsigned int GetVolSize(unsigned int iVolFile)
{
	FILE *volFile;
	unsigned char sVolFile[MAX_PATH];
	fpos_t pos;
	unsigned int volsize = 0;

	/* Get the VOL file name */
    sprintf(sVolFile, "%s\\vol.%i", sOutputPath, iVolFile);

	/* Open the VOL file for reading */
	if ((volFile = fopen(sVolFile, "rb")) != NULL) {
		/* Get the position of the cursor */
		fseek(volFile, 0, SEEK_END);
		fgetpos(volFile, &pos);

		volsize = (unsigned int)pos;

		/* Close files */
		fclose(volFile);
	}

	return volsize;
}


/***************************************************************************
** DecryptMessageSection
**
** Purpose: Decrypts the message section of a logic resource.
**
***************************************************************************/
void DecryptMessageSection(char *buffer, int len)
{
	const char key[] = "Avis Durgan";
	short offset;
	int i;
	unsigned char numMsg;

	offset = *(short *)buffer + 2;
	numMsg = buffer[offset];
	offset = offset + numMsg*2 + 3;

	for (i=0; i<len-offset; i++)
	{
		buffer[i+offset] ^= key[i % 11];		
	}
}

/***************************************************************************
** CleanUp
**
** Purpose: Deletes VOL and directory files in the specified output folder.
**
***************************************************************************/
void CleanUp()
{
	long hFind;
	struct _finddata_t finddata;
	char sFileSpec[MAX_PATH];
	char sFilename[MAX_PATH];
	int i;

	// Delete VOL files
	sprintf(sFileSpec, "%s\\vol.*", sOutputPath);
	hFind = _findfirst(sFileSpec, &finddata);
	if (hFind > -1) {
		/* Loop through the VOL files */
		do {
			if (!(finddata.attrib & _A_SUBDIR) ) {
				sprintf(sFilename, "%s\\%s", sOutputPath, finddata.name);
				remove(sFilename);
			}
		}
		while (_findnext(hFind, &finddata) > -1);
		_findclose(hFind);
	}

	// Delete directory files
	for (i=0; i<4; i++)
	{
		sprintf(sFilename, "%s\\%s", sOutputPath, ResFiles[i]);
		remove(sFilename);
	}
	
}

/***************************************************************************
** ShowUsage
**
** Purpose: Shows the command syntax.
**
***************************************************************************/
void ShowUsage() {

	printf("Usage   : buildvol [options] resource-path dir-path output-path\n");
	printf("Options : -v3   Write version 3 resource headers\n");
	printf("          -c	Compress resources (v3 only)\n");
	printf("          -b    Build one big vol file\n");

	/* Exit normally */		
	exit(0);

}
