
/***************************************************************************
** lzh.c
**
** This file has been borrowed from the ScummVM interpreter and has been
** modified to handle compression.
**
** Routines that deal with AGI version 3 specific features.
** The original LZW code is from DJJ, October 1989, p.86.
** It has been modified to handle AGI compression.
**
** (c) 1997  Lance Ewing
***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <windows.h>

#include "lzw.h"

#define MAXBITS			12
#define HASHING_SHIFT	(MAXBITS-8)
#define TABLE_SIZE		18041	/* strange number */
#define START_BITS		9

static int32 BITS, MAX_VALUE, MAX_CODE;
static int *code_value;
static uint32 *prefix_code;
static uint8 *append_character;
static uint8 *decode_stack;
static int32 input_bit_count = 0;	/* Number of bits in input bit buffer */
static uint32 input_bit_buffer = 0L;
static int32 output_bit_count = 0;	/* Number of bits in output bit buffer */
static uint32 output_bit_buffer = 0L;

static void initLZW() {
	decode_stack = (uint8 *)calloc(1, 8192);
	code_value=(int*)malloc(TABLE_SIZE * sizeof(int));
	prefix_code = (uint32 *)malloc(TABLE_SIZE * sizeof(uint32));
	append_character = (uint8 *)malloc(TABLE_SIZE * sizeof(uint8));
	input_bit_count = 0;	/* Number of bits in input bit buffer */
	input_bit_buffer = 0L;
	output_bit_count = 0;	/* Number of bits in output bit buffer */
	output_bit_buffer = 0L;
}

static void closeLZW() {
	free(decode_stack);
	free(code_value);
	free(prefix_code);
	free(append_character);
}

/***************************************************************************
** setBITS
**
** Purpose: To adjust the number of bits used to store codes to the value
** passed in.
***************************************************************************/
int setBITS(int32 value) {
	if (value == MAXBITS)
		return TRUE;

	BITS = value;
	MAX_VALUE = (1 << BITS) - 1;
	MAX_CODE = MAX_VALUE - 1;

	return FALSE;
}

/***************************************************************************
** decode_string
**
** Purpose: To return the string that the code taken from the input buffer
** represents. The string is returned as a stack, i.e. the characters are
** in reverse order.
***************************************************************************/
static uint8 *decode_string(uint8 *buffer, uint32 code) {
	uint32 i;

	for (i = 0; code > 255;) {
		*buffer++ = append_character[code];
		code = prefix_code[code];
		if (i++ >= 4000) {
			fprintf(stderr, "lzw: error in code expansion.\n");
			abort();
		}
	}
	*buffer = code;

	return buffer;
}

/***************************************************************************
** input_code
**
** Purpose: To return the next code from the input buffer.
***************************************************************************/
static uint32 input_code(uint8 **input) {
	uint32 r;

	while (input_bit_count <= 24) {
		input_bit_buffer |= (uint32) * (*input)++ << input_bit_count;
		input_bit_count += 8;
	}
	r = (input_bit_buffer & 0x7FFF) % (1 << BITS);
	input_bit_buffer >>= BITS;
	input_bit_count -= BITS;

	return r;
}

/***************************************************************************
** output_code (GM)
**
** Purpose: To send the specified code in the output buffer.
***************************************************************************/
static void output_code(uint8 **output, uint32 code)
{
	uint8 r;

	output_bit_buffer |= (uint32) code << output_bit_count;
	output_bit_count += BITS;

	while (output_bit_count >= 8)
	{
		r = (output_bit_buffer & 0xFF);
		*(*output)++ = r;
		output_bit_buffer >>= 8;
		output_bit_count -= 8;
	}

	/* Flush the output bit buffer if it's the last code */
	if (code == 257 && output_bit_count > 0)
	{
		r = (output_bit_buffer & 0xFF);
		*(*output)++ = r;
	}
}

/***************************************************************************
** expand
**
** Purpose: To uncompress the data contained in the input buffer and store
** the result in the output buffer. The fileLength parameter says how
** many bytes to uncompress. The compression itself is a form of LZW that
** adjusts the number of bits that it represents its codes in as it fills
** up the available codes. Two codes have special meaning:
**
**  code 256 = start over
**  code 257 = end of data
***************************************************************************/
void LZW_expand(uint8 *in, uint8 *out, int32 len) {
	int32 c, lzwnext, lzwnew, lzwold;
	uint8 *s, *end;

	initLZW();

	setBITS(START_BITS);	/* Starts at 9-bits */
	lzwnext = 257;		/* Next available code to define */

	end = (unsigned char *)((long)out + (long)len);

	lzwold = input_code(&in);	/* Read in the first code */
	c = lzwold;
	lzwnew = input_code(&in);

	while ((out < end) && (lzwnew != 0x101)) {
		if (lzwnew == 0x100) {
			/* Code to "start over" */
			lzwnext = 258;
			setBITS(START_BITS);
			lzwold = input_code(&in);
			c = lzwold;
			*out++ = (char)c;
			lzwnew = input_code(&in);
		} else {
			if (lzwnew >= lzwnext) {
				/* Handles special LZW scenario */
				*decode_stack = c;
				s = decode_string(decode_stack + 1, lzwold);
			} else
				s = decode_string(decode_stack, lzwnew);

			/* Reverse order of decoded string and
			 * store in out buffer
			 */
			c = *s;
			while (s >= decode_stack)
				*out++ = *s--;

			if (lzwnext > MAX_CODE)
				setBITS(BITS + 1);

			prefix_code[lzwnext] = lzwold;
			append_character[lzwnext] = c;
			lzwnext++;
			lzwold = lzwnew;

			lzwnew = input_code(&in);
		}
	}
	closeLZW();
}

/***************************************************************************
** find_match (GM)
**
** Purpose: To return the index of the specified string code/append 
** character combination in the string table.
***************************************************************************/
int find_match(uint32 hash_prefix, uint32 hash_character)
{
	int index;
	int offset;

	index = (hash_character << HASHING_SHIFT) ^ hash_prefix;
	if (index == 0)
		offset = 1;
	else
		offset = TABLE_SIZE - index;
	while (1)
	{
		if (code_value[index] == -1)
			return(index);
		if (prefix_code[index] == hash_prefix && 
			append_character[index] == hash_character)
			return(index);
		index -= offset;
		if (index < 0)
			index += TABLE_SIZE;
	}
}

/***************************************************************************
** compress (GM)
**
** Purpose: To compress the data contained in the input buffer and store
** the result in the output buffer. The fileLength parameter says how
** many bytes to compress. The compression itself is a form of LZW that
** adjusts the number of bits that it represents its codes in as it fills
** up the available codes. Two codes have special meaning:
**
**  code 256 = start over
**  code 257 = end of data
***************************************************************************/
uint32 LZW_compress(uint8 *in, uint8 *out, int32 len) {
	int32 lzwnext;
	uint32 c, string_code, index;
	uint8 *end, *start, *string_start;
	int i;

	initLZW();

	setBITS(START_BITS);	/* Starts at 9-bits */
	lzwnext = 258;		/* Next available code to define */

	start = out;
	end = (unsigned char *)((long)in + (long)len);

	/* Clear the string table */
	for (i=0; i < TABLE_SIZE; i++)	{
		code_value[i] = -1;	
	}

	output_code(&out, 256); /* First code is always 256 */
	string_code = *in++;	/* Get the first code */

	while (in < end) {
		c = *in++;
		index = find_match(string_code, c);

		if (code_value[index] != -1) {
			string_code = code_value[index];
		}
		else {
			if (lzwnext > MAX_CODE + 2) {
				if (setBITS(BITS + 1)) {
					/* Code to "start over" */
					output_code(&out, 256);
					/* Clear the string table */
					for (i=0; i < TABLE_SIZE; i++)	{
						code_value[i] = -1;	
					}
					setBITS(START_BITS);
					lzwnext = 258;
					in = string_start - 1;
					string_code = *in++;
					continue;
				}
			}
			code_value[index] = lzwnext++;
			prefix_code[index] = string_code;
			append_character[index] = c;
			output_code(&out, string_code);
			string_code = c;
			string_start = in;
		}
	}
	output_code(&out, string_code); /* output the last code */
	output_code(&out, 257); /* end of data code */
	
	closeLZW();

	/* return the length of the decompressed data */
	return out - start;
}
